#!/usr/bin/env python
# -*- coding: UTF-8 -*-
# Created by Roberto Preste
from typing import Union

import vcfpy

from mitoviz.constants import COLOR_MAPS
from mitoviz.utils import convert_hf, convert_nt, convert_plotly


class _Variant:
    """ Class storing a given variant, used for both linear and polar plots.

    Attributes:
        reference: reference allele of the variant
        position: position of the variant
        alternate: alternate allele of the variant
        hf: heteroplasmic fraction of the variant
    """

    def __init__(self,
                 reference: str,
                 position: int,
                 alternate: Union[str, vcfpy.Substitution],
                 hf: float):
        self.reference = reference
        self.position = position
        self.alternate = alternate
        self.hf = hf

    @property
    def _is_deletion(self) -> bool:
        """ Check whether the current variant refers to a deletion. """
        # e.g. ref CTG | alt C
        if isinstance(self.alternate, vcfpy.Substitution):
            return self.alternate.type == "DEL"
        return len(self.reference) > len(self.alternate)

    @property
    def _is_insertion(self) -> bool:
        """ Check whether the current variant refers to an insertion. """
        # e.g. ref C | alt CTG
        if isinstance(self.alternate, vcfpy.Substitution):
            return self.alternate.type == "INS"
        return len(self.reference) < len(self.alternate)

    @property
    def color(self) -> str:
        """ The color of the locus on which the variant is located. """
        for (start, stop), color in COLOR_MAPS.items():
            if self.position in range(start, stop + 1):
                return color

    @property
    def label(self) -> str:
        """ Create the variant label for deletions, insertions and SNPs. """
        alternate = (self.alternate.value
                     if isinstance(self.alternate, vcfpy.Substitution)
                     else self.alternate)
        if self._is_deletion:
            label = "{}d".format(int(self.position) + 1)
        elif self._is_insertion:
            label = "{}.{}".format(
                self.position, alternate[len(self.reference):]
            )
        else:
            label = "{}{}>{}".format(self.position, self.reference, alternate)

        return label

    @property
    def linear_x(self) -> int:
        """ The x position of the variant on the linear mt genome plot. """
        return self.position

    @property
    def linear_y(self) -> float:
        """ The y position of the variant on the linear mt genome plot. """
        return self.hf

    @property
    def polar_x(self) -> float:
        """ The x position of the variant on the polar mt genome plot. """
        return convert_nt(self.position)

    @property
    def polar_x_p(self) -> float:
        """ The x position of the variant on the polar plotly mt genome plot.
        """
        return convert_plotly(self.polar_x)

    @property
    def polar_y(self) -> float:
        """ The y position of the variant on the polar mt genome plot. """
        return 20 + convert_hf(self.hf)

    @property
    def strand(self) -> str:
        """ The mitochondrial strand on which the variant's locus is located.
        """
        if self.position in [*range(0, 576), *range(4331, 4400),
                             *range(5586, 5655), *range(5656, 5729),
                             *range(5729, 5761), *range(5760, 5826),
                             *range(5826, 5892), *range(7445, 7517),
                             *range(14148, 14673), *range(14673, 14742),
                             *range(15955, 16024), *range(16023, 16569)]:
            return "L"
        elif self.position in [*range(3304, 3306), *range(4400, 4401),
                               *range(5579, 5586), *range(5655, 5656),
                               *range(5891, 5903), *range(7514, 7517),
                               *range(8269, 8294), *range(8364, 8365),
                               *range(14742, 14746), *range(15953, 15955)]:
            return ""
        else:
            return "H"

    def __key(self):
        return self.reference, self.position, self.alternate, self.hf

    def __hash__(self):
        return hash(self.__key())

    def __eq__(self, other):
        if isinstance(other, _Variant):
            return self.__key() == other.__key()
        return NotImplemented

    def __repr__(self):
        return "{}(reference={}, position={}, alternate={}, hf={})".format(
            self.__class__.__name__, self.reference, self.position,
            self.alternate, self.hf
        )



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
# Created by Roberto Preste
from mitoviz.mitoviz import plot_base, plot_df, plot_table, plot_vcf  # noqa

__author__ = """Roberto Preste"""
__email__ = "robertopreste@gmail.com"
__version__ = '0.8.1'



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
# Created by Roberto Preste
import os
from typing import Optional

import matplotlib.pyplot as plt
import pandas as pd

from mitoviz.parsers import _DataFrameParser, _TabularParser, _VcfParser
from mitoviz.plot import (
    _plot_variants_polar, _plot_variants_linear, _plotly_variants_polar,
    _plotly_variants_linear, _plot_mito_polar, _plot_mito_linear,
    _plotly_mito_polar, _plotly_mito_linear
)
from mitoviz.utils import parse_path


def plot_base(linear: bool = False,
              save: bool = False,
              output: Optional[str] = None,
              legend: bool = False,
              split: bool = False,
              interactive: bool = False) -> None:
    """ Plot the base mitochondrial representation.

    Args:
        linear: show a linear plot rather than a polar one [default: False]
        save: if true, the final plot will be saved to a file [default: False]
        output: path of the output file where the plot will be saved
        legend: if true, add a legend for loci colors in the plot
            [default: False]
        split: if true, plot split H and L strands [default: True]
        interactive: if true, create an interactive version of the plot
            [default: False]
    """
    if linear:
        if interactive:
            plot_mito = _plotly_mito_linear
        else:
            plot_mito = _plot_mito_linear
    else:
        if interactive:
            plot_mito = _plotly_mito_polar
        else:
            plot_mito = _plot_mito_polar

    fig = plot_mito(legend, split)

    if save:
        dirname, name, ext = parse_path(output)
        if name == "":
            name = "base_mt"
        if interactive:
            fig.write_html(os.path.join(dirname, f"{name}.html"),
                           auto_open=False)
        else:
            plt.savefig(os.path.join(dirname, f"{name}{ext}"))
            plt.close()


def plot_vcf(in_vcf: str,
             linear: bool = False,
             sample: Optional[str] = None,
             save: bool = False,
             output: Optional[str] = None,
             labels: bool = False,
             legend: bool = False,
             split: bool = False,
             interactive: bool = False) -> None:
    """ Plot variants from the given VCF file.

    Args:
        in_vcf: path of the input VCF file
        linear: plot variants on a linear plot rather than a polar one
            [default: False]
        sample: specific sample to plot (defaults to all available samples)
        save: if true, the final plot will be saved to a file [default: False]
        output: path of the output file where the plot will be saved
        labels: if true, add a label for each variant shown [default: False]
        legend: if true, add a legend for loci colors in the plot
            [default: False]
        split: if true, plot split H and L strands [default: False]
        interactive: if true, create an interactive version of the plot
            [default: False]
    """
    vcf = _VcfParser(in_vcf)
    variants_per_sample = vcf.variants

    if linear:
        if interactive:
            plot_variants = _plotly_variants_linear
        else:
            plot_variants = _plot_variants_linear
    else:
        if interactive:
            plot_variants = _plotly_variants_polar
        else:
            plot_variants = _plot_variants_polar

    if sample:
        fig = plot_variants(sample, variants_per_sample[sample], labels,
                            legend, split)

        if save:
            dirname, name, ext = parse_path(output)
            if name == "":
                name = sample
            if interactive:
                fig.write_html(os.path.join(dirname, f"{name}.html"),
                               auto_open=False)
            else:
                plt.savefig(os.path.join(dirname, f"{name}{ext}"))
                plt.close()
    else:
        for i, (sample, variants) in enumerate(variants_per_sample.items(),
                                               start=1):
            fig = plot_variants(sample, variants, labels, legend, split)

            if save:
                dirname, name, ext = parse_path(output)
                if name == "":
                    name = sample
                    if interactive:
                        fig.write_html(os.path.join(dirname, f"{name}.html"),
                                       auto_open=False)
                    else:
                        plt.savefig(os.path.join(dirname, f"{name}{ext}"))
                        plt.close()
                elif len(variants_per_sample) == 1:
                    if interactive:
                        fig.write_html(os.path.join(dirname, f"{name}.html"),
                                       auto_open=False)
                    else:
                        plt.savefig(os.path.join(dirname, f"{name}{ext}"))
                        plt.close()
                else:
                    if interactive:
                        fig.write_html(os.path.join(dirname,
                                                    f"{name}_{i}.html"),
                                       auto_open=False)
                    else:
                        plt.savefig(os.path.join(dirname, f"{name}_{i}{ext}"))
                        plt.close()


def plot_df(in_df: pd.DataFrame,
            linear: bool = False,
            sample: Optional[str] = None,
            save: bool = False,
            output: Optional[str] = None,
            labels: bool = False,
            legend: bool = False,
            split: bool = False,
            interactive: bool = False,
            pos_col: str = "POS",
            ref_col: str = "REF",
            alt_col: str = "ALT",
            sample_col: str = "SAMPLE",
            hf_col: str = "HF") -> None:
    """ Plot variant from the given pandas DataFrame.

    Args:
        in_df: input pandas DataFrame
        linear: plot variants on a linear plot rather than a polar one
            [default: False]
        sample: specific sample to plot (defaults to all available samples)
        save: if true, the final plot will be saved to a file [default: False]
        output: path of the output file where the plot will be saved
        labels: if true, add a label for each variant shown [default: False]
        legend: if true, add a legend for loci colors in the plot
            [default: False]
        split: if true, plot split H and L strands [default: False]
        interactive: if true, create an interactive version of the plot
            [default: False]
        pos_col: column name for the variant position
        ref_col: column name for the variant reference allele
        alt_col: column name for the variant alternate allele
        sample_col: column name for the variant sample
        hf_col: column name for the variant heteroplasmic fraction
    """
    df = _DataFrameParser(in_df,
                          pos_col=pos_col,
                          ref_col=ref_col,
                          alt_col=alt_col,
                          sample_col=sample_col,
                          hf_col=hf_col)
    variants_per_sample = df.variants

    if linear:
        if interactive:
            plot_variants = _plotly_variants_linear
        else:
            plot_variants = _plot_variants_linear
    else:
        if interactive:
            plot_variants = _plotly_variants_polar
        else:
            plot_variants = _plot_variants_polar

    if sample:
        fig = plot_variants(sample, variants_per_sample[sample], labels,
                            legend, split)

        if save:
            dirname, name, ext = parse_path(output)
            if name == "":
                name = sample
            if interactive:
                fig.write_html(os.path.join(dirname, f"{name}.html"),
                               auto_open=False)
            else:
                plt.savefig(os.path.join(dirname, f"{name}{ext}"))
                plt.close()
    else:
        for i, (sample, variants) in enumerate(variants_per_sample.items(),
                                               start=1):
            fig = plot_variants(sample, variants, labels, legend, split)

            if save:
                dirname, name, ext = parse_path(output)
                if name == "":
                    name = sample
                    if interactive:
                        fig.write_html(os.path.join(dirname, f"{name}.html"),
                                       auto_open=False)
                    else:
                        plt.savefig(os.path.join(dirname, f"{name}{ext}"))
                        plt.close()
                elif len(variants_per_sample) == 1:
                    if interactive:
                        fig.write_html(os.path.join(dirname, f"{name}.html"),
                                       auto_open=False)
                    else:
                        plt.savefig(os.path.join(dirname, f"{name}{ext}"))
                        plt.close()
                else:
                    if interactive:
                        fig.write_html(os.path.join(dirname,
                                                    f"{name}_{i}.html"),
                                       auto_open=False)
                    else:
                        plt.savefig(os.path.join(dirname, f"{name}_{i}{ext}"))
                        plt.close()


def plot_table(in_table: str,
               sep: str = ",",
               linear: bool = False,
               sample: Optional[str] = None,
               save: bool = False,
               output: Optional[str] = None,
               labels: bool = False,
               legend: bool = False,
               split: bool = False,
               interactive: bool = False,
               pos_col: str = "POS",
               ref_col: str = "REF",
               alt_col: str = "ALT",
               sample_col: str = "SAMPLE",
               hf_col: str = "HF",
               **kwargs) -> None:
    """ Plot variants from the given tabular file.

    Args:
        in_table: path of the input tabular file
        sep: column delimiter used [default: ',']
        linear: plot variants on a linear plot rather than a polar one
            [default: False]
        sample: specific sample to plot (defaults to all available samples)
        save: if true, the final plot will be saved to a file [default: False]
        output: path of the output file where the plot will be saved
        labels: if true, add a label for each variant shown [default: False]
        legend: if true, add a legend for loci colors in the plot
            [default: False]
        split: if true, plot split H and L strands [default: False]
        interactive: if true, create an interactive version of the plot
            [default: False]
        pos_col: column name for the variant position
        ref_col: column name for the variant reference allele
        alt_col: column name for the variant alternate allele
        sample_col: column name for the variant sample
        hf_col: column name for the variant heteroplasmic fraction
        **kwargs: additional arguments passed to pandas.read_table()
    """
    table = _TabularParser(in_table, sep=sep, **kwargs)
    plot_df(table.df,
            linear=linear,
            sample=sample,
            save=save,
            output=output,
            labels=labels,
            legend=legend,
            split=split,
            interactive=interactive,
            pos_col=pos_col,
            ref_col=ref_col,
            alt_col=alt_col,
            sample_col=sample_col,
            hf_col=hf_col)



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
# Created by Roberto Preste
from typing import List

import matplotlib.patches as mpatches
import matplotlib.pyplot as plt
import numpy as np
import plotly.graph_objects as go

from mitoviz.constants import COLORS, NAMES
from mitoviz.locus import _LinearLocus, _PolarLocus, _PolarSplitLocus
from mitoviz.variant import _Variant


def _label_variant(ax: plt.axes, variant: _Variant, linear: bool = False):
    """ Annotate each variant with a label in the polar or linear plot.

    Args:
        ax: axis element on which the label will be added
        variant: variant to annotate
        linear: whether the resulting is a linear or polar plot
            [default: False]
    """
    if linear:
        ax.annotate(variant.label,
                    xy=(variant.linear_x, variant.linear_y + 0.02),
                    xytext=(variant.linear_x, variant.linear_y + 0.02),
                    ha="center", va="bottom",
                    bbox=dict(facecolor="w",
                              alpha=0.8,
                              boxstyle="round"))
    else:
        ax.annotate(variant.label,
                    xy=(variant.polar_x, variant.polar_y),
                    xytext=(variant.polar_x, variant.polar_y),
                    textcoords="offset pixels",
                    ha="center", va="bottom",
                    bbox=dict(facecolor="w",
                              alpha=0.8,
                              boxstyle="round"))


def _legend_patches() -> List[mpatches.Patch]:
    """ Return a list of mpatches.Patch to create the loci legend. """
    cds = mpatches.Patch(color=COLORS["cds"], label="Coding")
    reg = mpatches.Patch(color=COLORS["reg"], label="Regulatory")
    rrna = mpatches.Patch(color=COLORS["rrna"], label="rRNA")
    trna = mpatches.Patch(color=COLORS["trna"], label="tRNA")
    nc = mpatches.Patch(color=COLORS["nc"], label="Non Coding")
    return [cds, reg, rrna, trna, nc]


def _legend_plotly_polar() -> List[go.Scatterpolar]:
    """ Return a list of plotly Scatterpolar elements to create the loci
    legend on a polar plot. """
    cds = go.Scatterpolar(r=[None], theta=[None], mode="markers",
                          marker=dict(size=10, color="#2e8b57"),
                          showlegend=True, name="Coding")
    reg = go.Scatterpolar(r=[None], theta=[None], mode="markers",
                          marker=dict(size=10, color="#ffa500"),
                          showlegend=True, name="Regulatory")
    rrna = go.Scatterpolar(r=[None], theta=[None], mode="markers",
                           marker=dict(size=10, color="#cd5c5c"),
                           showlegend=True, name="rRNA")
    trna = go.Scatterpolar(r=[None], theta=[None], mode="markers",
                           marker=dict(size=10, color="#4169e1"),
                           showlegend=True, name="tRNA")
    nc = go.Scatterpolar(r=[None], theta=[None], mode="markers",
                         marker=dict(size=10, color="grey"),
                         showlegend=True, name="Non Coding")
    return [cds, reg, rrna, trna, nc]


def _legend_plotly_linear() -> List[go.Scatterpolar]:
    """ Return a list of plotly Scatter elements to create the loci legend
    on a linear plot. """
    cds = go.Scatter(x=[None], y=[None], mode="markers",
                     marker=dict(size=10, color="#2e8b57"),
                     showlegend=True, name="Coding")
    reg = go.Scatter(x=[None], y=[None], mode="markers",
                     marker=dict(size=10, color="#ffa500"),
                     showlegend=True, name="Regulatory")
    rrna = go.Scatter(x=[None], y=[None], mode="markers",
                      marker=dict(size=10, color="#cd5c5c"),
                      showlegend=True, name="rRNA")
    trna = go.Scatter(x=[None], y=[None], mode="markers",
                      marker=dict(size=10, color="#4169e1"),
                      showlegend=True, name="tRNA")
    nc = go.Scatter(x=[None], y=[None], mode="markers",
                    marker=dict(size=10, color="grey"),
                    showlegend=True, name="Non Coding")
    return [cds, reg, rrna, trna, nc]


def _plot_mito_linear(legend: bool = False,
                      split: bool = False):
    """ Return an axes object with the base mt genome linear plot.

    Args:
        legend: add a legend for loci colors in the plot [default: False]
        split: plot split H and L strands [default: False]
    """
    fig = plt.figure(figsize=(20, 10), dpi=300)
    ax = fig.add_subplot(111)
    names = NAMES + ["DLOOP"]
    loci = [_LinearLocus(name=name, index=index)
            for index, name in enumerate(names)]

    if split:
        h_loci = [locus for locus in loci if locus.strand == "H"]
        l_loci = [locus for locus in loci if locus.strand == "L"]
        nc_loci = [locus for locus in loci if locus.strand == ""]
        ax.broken_barh([(locus.start, locus.width) for locus in h_loci],
                       (-0.05, 0.05),
                       facecolors=[locus.color for locus in h_loci])
        ax.broken_barh([(locus.start, locus.width) for locus in l_loci],
                       (-0.1, 0.05),
                       facecolors=[locus.color for locus in l_loci])
        ax.broken_barh([(locus.start, locus.width) for locus in nc_loci],
                       (-0.1, 0.1),
                       facecolors=[locus.color for locus in nc_loci])
    else:
        ax.broken_barh([(locus.start, locus.width) for locus in loci],
                       (-0.1, 0.1),
                       facecolors=[f"{locus.color}" for locus in loci])
    ax.set_xticks([0, 4000, 8000, 12000, 16000])
    ax.set_yticks([0.0, 0.25, 0.50, 0.75, 1.00])
    ax.set_ylim(-0.15, 1.10)
    ax.spines["top"].set_visible(False)
    ax.spines["right"].set_visible(False)

    for locus in loci:
        if locus.name == "OLR":
            ax.annotate(locus.name,
                        xy=(locus.text_x, -0.11),
                        ha="center",
                        fontsize=4)
            continue
        if locus.loc_type != "nc":
            ax.annotate(locus.name,
                        xy=(locus.text_x, locus.text_y),
                        ha="center",
                        fontsize=5)

    if legend:
        handles = _legend_patches()
        plt.legend(handles=handles, loc="upper right")
        ax.set_xlim([-500, 18000])

    return fig, ax


def _plotly_mito_linear(legend: bool = False,
                        split: bool = False) -> go.Figure:
    """Return a plotly figure with the base mt genome linear plot.

    Args:
        legend: add a legend for loci colors in the plot [default: False]
        split: plot split H and L strand [default: False]
    """
    names = NAMES + ["DLOOP"]
    loci = [_LinearLocus(name=name, index=index)
            for index, name in enumerate(names)]

    if split:
        shapes = []
        for locus in loci:
            if locus.strand == "H":
                y0 = -0.1
                y1 = 0.0
            elif locus.strand == "L":
                y0 = -0.2
                y1 = -0.1
            else:
                y0 = -0.2
                y1 = 0.0
            shape = dict(type="rect",
                         x0=locus.start,
                         y0=y0,
                         x1=(locus.start + locus.width),
                         y1=y1,
                         fillcolor=locus.color,
                         line_color=locus.color)
            shapes.append(shape)
    else:
        shapes = [dict(type="rect",
                       x0=locus.start,
                       y0=-0.2,
                       x1=(locus.start + locus.width),
                       y1=0.0,
                       fillcolor=locus.color,
                       line_color=locus.color) for locus in loci]

    scatters = go.Scatter(
        x=[locus.text_x for locus in loci],
        y=[-0.1 for _ in loci],
        text=[locus.name for locus in loci],
        hovertemplate="%{text}<extra></extra>",
        mode="markers",
        marker=dict(symbol="square", opacity=0),
        showlegend=False,
    )
    fig = go.Figure()
    fig.add_trace(scatters)

    if legend:
        legend = _legend_plotly_linear()
        for el in legend:
            fig.add_trace(el)

    fig.update_layout(width=1000, height=800, template="none",
                      legend=dict(x=0.0, y=1.0,
                                  xanchor="right", yanchor="middle"),
                      xaxis=dict(showgrid=False,
                                 showline=False,
                                 zeroline=False),
                      yaxis=dict(showgrid=False,
                                 showline=False,
                                 zeroline=False),
                      yaxis_range=[-0.3, 1.2],
                      hovermode="closest",
                      shapes=shapes)
    fig.update_xaxes(tickvals=[0, 4000, 8000, 12000, 16000])
    fig.update_yaxes(tickvals=[0, 0.25, 0.5, 0.75, 1.0])

    return fig


def _plot_mito_polar(legend: bool = False,
                     split: bool = False):
    """ Return an axes object with the base mt genome polar plot.

    Args:
        legend: add a legend for loci colors in the plot [default: False]
        split: plot split H and L strands [default: False]
    """
    fig = plt.figure(figsize=(20, 10), dpi=300)
    ax = fig.add_subplot(111, polar=True)
    if split:
        names = NAMES + [""]
        loci = [_PolarSplitLocus(name=name, index=index)
                for index, name in enumerate(names)]
        radii = [el.radius for el in loci]
        bottoms = [el.bottom for el in loci]
    else:
        loci = [_PolarLocus(name=name, index=index)
                for index, name in enumerate(NAMES)]
        radii = 5.0
        bottoms = 20.0
    thetas = [el.theta for el in loci]
    widths = [el.width for el in loci]

    bars = plt.bar(thetas, radii, width=widths, bottom=bottoms)

    # Add black borders, useful for split plots
    ax.bar(0, 0.1, 2 * np.pi, 19.9, facecolor="black")
    ax.bar(0, 0.1, 2 * np.pi, 25.0, facecolor="black")

    for locus, bar in zip(loci, bars):
        bar.set_facecolor(locus.color)
        if locus.name == "OLR":
            ax.annotate(locus.name,
                        xy=(locus.theta, locus.text_y),
                        ha=locus.text_ha, va=locus.text_va,
                        fontsize=6)
            continue
        if locus.loc_type != "nc":
            ax.annotate(locus.name,
                        xy=(locus.theta, locus.text_y),
                        ha=locus.text_ha, va=locus.text_va)

    ax.set_xticklabels([])
    ax.set_yticklabels([])
    ax.xaxis.grid(False)
    ax.yaxis.grid(False)
    ax.spines["polar"].set_visible(False)
    ax.set_theta_zero_location("N")

    if legend:
        handles = _legend_patches()
        plt.legend(handles=handles, loc="center")

    return fig, ax


def _plotly_mito_polar(legend: bool = False,
                       split: bool = False) -> go.Figure:
    """ Return a plotly figure with the base mt genome polar plot.

    Args:
        legend: add a legend for loci colors in the plot [default: False]
        split: plot split H and L strand [default: False]
    """
    fig = go.Figure()

    if split:
        names = NAMES + [""]
        loci = [_PolarSplitLocus(name=name, index=index)
                for index, name in enumerate(names)]
        radii = [el.radius for el in loci]
        bottoms = [el.bottom for el in loci]
    else:
        loci = [_PolarLocus(name=name, index=index)
                for index, name in enumerate(NAMES)]
        radii = [5.0 for _ in loci]
        bottoms = [20.0 for _ in loci]
    names = [el.name for el in loci]
    thetas = [el.theta_p for el in loci]
    widths = [el.width_p for el in loci]
    colors = [el.color for el in loci]

    mito_trace = go.Barpolar(r=radii, theta=thetas, width=widths, base=bottoms,
                             marker_color=colors, meta=names,
                             hovertemplate="%{meta}<extra></extra>",
                             showlegend=False)

    fig.add_trace(mito_trace)

    border_trace = go.Barpolar(r=[0.1, 0.1], theta=[0, 0], width=[360, 360],
                               base=[19.9, 25.0], marker_color="black",
                               hoverinfo="none", showlegend=False)

    fig.add_trace(border_trace)

    if legend:
        legend = _legend_plotly_polar()
        for el in legend:
            fig.add_trace(el)

    fig.update_layout(width=900, height=800, template="none",
                      legend=dict(x=0.5, y=0.5,
                                  xanchor="center", yanchor="middle"),
                      polar=dict(
                          radialaxis=dict(showticklabels=False, ticks="",
                                          showgrid=False, showline=False),
                          angularaxis=dict(rotation=90, showticklabels=False,
                                           ticks="", showgrid=False,
                                           showline=False)))

    return fig


def _plot_variants_polar(sample: str,
                         variants: List[_Variant],
                         labels: bool = False,
                         legend: bool = False,
                         split: bool = False):
    """ Plot variants available in the given list onto a polar plot.

    Args:
        sample: sample name, used for the plot title
        variants: list of _PolarVariant instances to plot
        labels: add a label for each variant shown [default: False]
        legend: add a legend for loci colors in the plot [default: False]
        split: plot split H and L strands [default: False]
    """
    fig, ax = _plot_mito_polar(legend, split)

    for variant in variants:
        ax.scatter(variant.polar_x, variant.polar_y,
                   c="black", s=20, zorder=20)
        if labels:
            _label_variant(ax, variant, linear=False)

    ax.set_title(sample)

    return fig, ax


def _plotly_variants_polar(sample: str,
                           variants: List[_Variant],
                           labels: bool = False,
                           legend: bool = False,
                           split: bool = False) -> go.Figure:
    """ Plot variants available in the given list onto a plotly polar plot.

    Args:
        sample: sample name, used for the plot title
        variants: list of _PolarVariant instances to plot
        labels: add a label for each variant shown [default: False]
        legend: add a legend for loci colors in the plot [default: False]
        split: plot split H and L strands [default: False]
    """
    fig = _plotly_mito_polar(legend, split)

    radii = [el.polar_y for el in variants]
    theta = [el.polar_x_p for el in variants]
    meta = [el.label for el in variants]

    var_trace = go.Scatterpolar(r=radii, theta=theta, mode="markers",
                                marker=dict(color="black"), meta=meta,
                                hovertemplate="%{meta}<extra></extra>",
                                showlegend=False)

    fig.add_trace(var_trace)
    fig.update_layout(title=sample)

    # Returning go.Figure to allow saving the html image
    return fig


def _plot_variants_linear(sample: str,
                          variants: List[_Variant],
                          labels: bool = False,
                          legend: bool = False,
                          split: bool = False):
    """ Plot variant available in a given list onto a linear plot.

    Args:
        sample: sample name, used for the plot title
        variants: list of _LinearVariant instances to plot
        labels: add a label for each variant shown [default: False]
        legend: add a legend for loci colors in the plot [default: False]
        split: plot split H and L strands [default: False]
    """
    fig, ax = _plot_mito_linear(legend, split)

    for variant in variants:
        if split:
            bottom = -0.05 if variant.strand == "L" else 0.0
        else:
            bottom = 0.0
        marker, stem, base = plt.stem([variant.linear_x], [variant.linear_y],
                                      "-.", bottom=bottom,
                                      use_line_collection=True)
        plt.setp(marker, "color", variant.color)
        plt.setp(stem, "color", variant.color)
        plt.setp(base, "linestyle", "None")

        if labels:
            _label_variant(ax, variant, linear=True)

    ax.set_title(sample)

    return fig, ax


def _plotly_variants_linear(sample: str,
                            variants: List[_Variant],
                            labels: bool = False,
                            legend: bool = False,
                            split: bool = False) -> go.Figure:
    """ Plot variant available in a given list onto a plotly linear plot.

    Args:
        sample: sample name, used for the plot title
        variants: list of _LinearVariant instances to plot
        labels: add a label for each variant shown [default: False]
        legend: add a legend for loci colors in the plot [default: False]
        split: plot split H and L strands [default: False]
    """
    fig = _plotly_mito_linear(legend, split)

    xs = [variant.linear_x for variant in variants]
    ys = [variant.linear_y for variant in variants]
    meta = [el.label for el in variants]

    var_trace = go.Scatter(x=xs, y=ys, mode="markers",
                           marker=dict(color="black"), meta=meta,
                           hovertemplate="%{meta}<extra></extra>",
                           showlegend=False)
    lines = [dict(type="line",
                  xref="x",
                  yref="y",
                  x0=variant.linear_x,
                  y0=-0.1,
                  x1=variant.linear_x,
                  y1=(variant.linear_y - 0.006),
                  layer="below",
                  line=dict(color="grey", width=1))
             for variant in variants]

    fig.add_trace(var_trace)
    base_shapes = list(fig.layout.shapes)
    fig.update_layout(shapes=[*base_shapes, *lines], title=sample)

    # Returning go.Figure to allow saving the html image
    return fig



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
# Created by Roberto Preste
import os
from typing import Optional, Tuple


def convert_nt(nt: int) -> float:
    """ Convert a nucleotide position/width to a relative value.

    Used to calculate relative values required for polar plots.

    Args:
        nt: position/width in nucleotides

    Returns:
        relative value
    """
    return (nt * 6.29) / 16569


def convert_hf(hf: float) -> float:
    """ Convert a variant's HF value from the [0.0, 1.0] range to [0.0, 5.0].

    Used to plot the variant on a polar plot.

    Args:
        hf: heteroplasmic fraction of the variant

    Returns:
        adjusted HF value
    """
    return hf * 5


def convert_plotly(value: float) -> float:
    """ Convert the given element to the proper value used in plotly polar
    plots.

    Works with _Variant.polar_x, _PolarLocus.theta, _PolarLocus.width, using
    the somehow magic number 57.1.

    Args:
        value: input value to convert

    Returns:
        adjusted value
    """
    return value * 57.1


def parse_path(path: Optional[str]) -> Tuple[str, str, str]:
    """ Parse the given path into directory name, file name and extension.

    Args:
        path: output path to parse; if None, returns the current directory
            and an empty file name

    Returns:
        tuple with directory name, file name and extension
    """
    if path:
        dirname = os.path.dirname(path)
        filename = os.path.basename(path)
        name, ext = os.path.splitext(filename)
    else:
        dirname = os.getcwd()
        name = ""
        ext = ".png"

    return dirname, name, ext



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
# Created by Roberto Preste
from typing import Tuple

from mitoviz.classes.base_locus import _BaseLocus
from mitoviz.constants import (
    NT_LENGTHS, STARTS, TEXT_HA, TEXT_VA, TEXT_Y, TYPES
)
from mitoviz.utils import convert_nt, convert_plotly


class _PolarLocus(_BaseLocus):
    """ Class referring to a single mt locus, used in polar plots.

    Attributes:
        name: name of the locus
        index: index of the locus in the mt genome (dloop = 0, tf = 1, etc.)
    """
    _nt_lengths = NT_LENGTHS
    _text_ha = TEXT_HA
    _text_va = TEXT_VA
    _text_y = TEXT_Y
    _types = TYPES

    def __init__(self, name: str, index: int):
        super().__init__(name=name, index=index)

    @property
    def loc_type(self) -> str:
        """ The locus type (regulatory, coding, rRNA, tRNA, non-coding). """
        return self._types[self.index]

    @property
    def text_ha(self) -> str:
        """ The horizontal alignment for the text label. """
        return self._text_ha[self.index]

    @property
    def text_va(self) -> str:
        """ The vertical alignment for the text label. """
        return self._text_va[self.index]

    @property
    def text_y(self) -> float:
        """ The y position for the text label. """
        return self._text_y[self.index]

    @property
    def theta(self) -> float:
        """ The position at which the locus will be plotted. """
        if self.index == 0:
            return 0.0
        elif self.index == 1:
            return convert_nt(self._nt_lengths[0])/2 + self.width/2
        return (convert_nt(self._nt_lengths[0])/2 + self.width/2
                + sum(map(convert_nt, self._nt_lengths[1:self.index])))

    @property
    def theta_p(self) -> float:
        """ The position at which the locus will be plotted when using plotly.
        """
        return convert_plotly(self.theta)

    @property
    def width(self) -> float:
        """ The relative width of the locus from its length in
        nucleotides. """
        return convert_nt(self._nt_lengths[self.index])

    @property
    def width_p(self) -> float:
        """ The relative width of the locus from its length in
        nucleotides when using plotly. """
        return convert_plotly(self.width)


class _PolarSplitLocus(_PolarLocus):
    """ Class referring to a single mt locus, used in polar plots with
    split strands.

    Attributes:
        name: name of the locus
        index: index of the locus in the mt genome (dloop = 0, tf = 1, etc.)
    """
    _nt_lengths = NT_LENGTHS + [546]
    _text_ha = TEXT_HA + ["center"]
    _text_va = TEXT_VA + ["center"]
    _text_y = TEXT_Y + [19.2]
    _types = TYPES + ["reg"]

    def __init__(self, name: str, index: int):
        super().__init__(name=name, index=index)

    @property
    def bottom(self) -> float:
        """ The internal position of the locus, used to plot the locus on
        the H/L strand. """
        if self.strand == "H":
            return 22.5
        return 20.0

    @property
    def radius(self) -> float:
        """ The radius of the locus, used to plot the locus on the H/L
        strand. """
        if self.strand == "":  # non coding
            return 5.0
        return 2.5


class _LinearLocus(_BaseLocus):
    """ Class referring to a single mt locus, used in linear plots.

    Attributes:
        name: name of the locus
        index: index of the locus in the mt genome (dloop = 0, tf = 1, etc.)
    """
    _nt_lengths = [576] + NT_LENGTHS[1:] + [546]
    _starts = STARTS
    _types = TYPES + ["reg"]

    def __init__(self, name: str, index: int):
        super().__init__(name=name, index=index)

    @property
    def height(self) -> Tuple[float, float]:
        """ The vertical position of the locus, if plotting split strands. """
        if self.strand == "H":
            return (-0.05, 0.05)
        elif self.strand == "L":
            return (-0.1, 0.05)
        return (-0.1, 0.1)

    @property
    def loc_type(self) -> str:
        """ The locus type (regulatory, coding, rRNA, tRNA). """
        return self._types[self.index]

    @property
    def start(self) -> int:
        """ Position on the mt genome where the locus starts. """
        return self._starts[self.index]

    @property
    def text_x(self) -> float:
        """ The x position for the text label. """
        return self.start + (self.width / 2)

    @property
    def text_y(self) -> float:
        """ The y position for the text label. """
        if self.name in ["TQ", "TA", "TC", "TD", "TK", "TS2", "TP"]:
            return -0.13
        return -0.12

    @property
    def width(self) -> float:
        """ The width of the locus in nucleotides. """
        return self._nt_lengths[self.index]



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
# Created by Roberto Preste
from setuptools import setup, find_packages

with open("README.rst") as readme_file:
    readme = readme_file.read()

with open("HISTORY.rst") as history_file:
    history = history_file.read()

requirements = ["Click>=7.0", "numpy", "pandas", "matplotlib", "pysam",
                "vcfpy", "plotly"]

setup_requirements = ["pytest-runner", ]

test_requirements = ["pytest", "opencv-python"]

setup(  # pragma: no cover
    author="Roberto Preste",
    author_email="robertopreste@gmail.com",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
    ],
    description="Plot variants on the human mitochondrial genome.",
    entry_points={
        "console_scripts": [
            "mitoviz=mitoviz.cli.mitoviz_plot:main",
            "mitoviz-base=mitoviz.cli.mitoviz_base:main"
        ],
    },
    install_requires=requirements,
    license="MIT license",
    long_description=readme + "\n\n" + history,
    long_description_content_type="text/x-rst",
    include_package_data=True,
    keywords="mitoviz",
    name="mitoviz",
    packages=find_packages(include=["mitoviz"],
                           exclude=["ci", "mitoviz/tests/imgs"]),
    setup_requires=setup_requirements,
    test_suite="tests",
    tests_require=test_requirements,
    url="https://github.com/robertopreste/mitoviz",
    version='0.8.1',
    zip_safe=False,
)



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
# Created by Roberto Preste
from collections import defaultdict
from typing import List

import pandas as pd
import vcfpy

from mitoviz.variant import _Variant


class _VcfParser:
    """ Class to read and parse the given VCF file.

    Attributes
        vcf_in: path of the input VCF file
    """

    def __init__(self, vcf_in: str):
        self.vcf_in = vcf_in
        self._variants = defaultdict(list)
        self._reader = vcfpy.Reader.from_path(vcf_in)
        self.parse_variants()

    @property
    def samples(self) -> List[str]:
        return self._reader.header.samples.names

    @property
    def variants(self) -> dict:
        return self._variants

    @staticmethod
    def parse_call(call: vcfpy.Call, i: int) -> float:
        """ Parse the vcfpy.Call to get the i-th value for HF; if not present,
        return 0.5.

        Args:
            call: input call to parse
            i: i-th element of HF to get

        Returns:
            either the required i-th element of HF or 0.5
        """
        hf_list = call.data.get("HF", [])
        if hf_list:
            return hf_list[i]
        return 0.5

    def parse_variants(self):
        """ Read the variants from the input VCF file and parse them in the
        required format.

        Variants are stored in a per-sample fashion, in the self.variants
        dictionary.
        """
        for record in self._reader:
            if self.samples:  # sample name is specified
                for sample, call in record.call_for_sample.items():
                    variants = [_Variant(record.REF, record.POS, alt,
                                         self.parse_call(call, i))
                                for i, alt in enumerate(record.ALT)]
                    self._variants[sample].extend(variants)
            else:  # single sample without name
                sample = "MITOVIZ001"
                variants = [_Variant(record.REF, record.POS, alt, 0.5)
                            for alt in record.ALT]
                self._variants[sample].extend(variants)

    def __repr__(self):
        return "{}(vcf_in={})".format(
            self.__class__.__name__, self.vcf_in
        )


class _DataFrameParser:
    """ Class to read and parse a given pandas DataFrame.

    Attributes:
        df_in: path of the input pandas DataFrame
        pos_col: column name for the variant position
        ref_col: column name for the variant reference allele
        alt_col: column name for the variant alternate allele
        sample_col: column name for the variant sample
        hf_col: column name for the variant heteroplasmic fraction
    """

    def __init__(self,
                 df_in: pd.DataFrame,
                 pos_col: str = "POS",
                 ref_col: str = "REF",
                 alt_col: str = "ALT",
                 sample_col: str = "SAMPLE",
                 hf_col: str = "HF"):
        self.df_in = df_in
        self.pos_col = pos_col
        self.ref_col = ref_col
        self.alt_col = alt_col
        self.sample_col = sample_col
        self.hf_col = hf_col
        self._variants = defaultdict(list)
        self.parse_variants()

    @property
    def has_hf(self) -> bool:
        """ Return whether the HF column is available in the DataFrame. """
        return self.hf_col in self.df_in.columns

    @property
    def has_sample(self) -> bool:
        """ Return whether the SAMPLE column is available in the DataFrame. """
        return self.sample_col in self.df_in.columns

    @property
    def variants(self) -> dict:
        return self._variants

    def parse_variants(self):
        """ Read the variants from the input DataFrame and parse them in the
        required format.

        Variants are stored in a per-sample fashion, in the self.variants
        dictionary.
        """
        if self.has_hf:
            self.df_in[self.hf_col] = self.df_in[self.hf_col].astype(float)
        for record in self.df_in.itertuples():
            rec = record._asdict()
            sample = rec[self.sample_col] if self.has_sample else "MITOVIZ001"
            hf = rec[self.hf_col] if self.has_hf else 0.5
            variant = _Variant(rec[self.ref_col], rec[self.pos_col],
                               rec[self.alt_col], hf)
            self._variants[sample].append(variant)

    def __repr__(self):
        return ("{}(pos_col={}, ref_col={}, alt_col={}, "
                "sample_col={}, hf_col={})").format(
            self.__class__.__name__, self.pos_col, self.ref_col,
            self.alt_col, self.sample_col, self.hf_col
        )


class _TabularParser:
    """ Class to read and parse a given tabular generic file.

    Attributes:
        table_in: path of the input tabular file
        sep: column delimiter used
        **kwargs: additional arguments passed to pandas.read_table()
    """

    def __init__(self,
                 table_in: str,
                 sep: str = ",",
                 **kwargs):
        self.table_in = table_in
        self.sep = sep
        self.kwargs = kwargs

    @property
    def df(self) -> pd.DataFrame:
        try:
            df_in = pd.read_table(self.table_in, sep=self.sep, engine="python",
                                  **self.kwargs)
        except TypeError as e:
            raise TypeError(e)
        return df_in

    def __repr__(self):
        return "{}(table_in={}, sep={!r})".format(self.__class__.__name__,
                                                  self.table_in, self.sep)
